from typing import Optional
import logging
import enum
from functools import partial
from time import sleep
from . import nm
from . import settings
from . import storage
from .state_machine import BaseState
from .app import Application
from .server import ConfiguredServer as Server, Protocol
from .utils import run_in_background_thread, translated_property

logger = logging.getLogger(__name__)


class StatusImage(enum.Enum):
    # The value is the image filename.
    DEFAULT = 'desktop-default.png'
    CONNECTING = 'desktop-connecting.png'
    CONNECTED = 'desktop-connected.png'
    NOT_CONNECTED = 'desktop-not-connected.png'

    @property
    def path(self):
        return settings.IMAGE_PREFIX + self.value


class NetworkState(BaseState):
    """
    Base class for all interface states.
    """

    status_label = translated_property("Connection state unknown")
    status_image = StatusImage.DEFAULT

    def start_new_connection(self,
                             app: Application,
                             server: Server,
                             ) -> 'NetworkState':
        return connect_initial(app)

    def set_connecting(self, app: Application) -> 'NetworkState':
        return ConnectingState()

    def set_connected(self, app: Application) -> 'NetworkState':
        return ConnectedState()

    def set_disconnected(self, app: Application) -> 'NetworkState':
        return DisconnectedState()

    def set_unknown(self, app: Application) -> 'NetworkState':
        return enter_unknown_state(app)

    def set_error(self, app: Application, message: Optional[str] = None) -> 'NetworkState':
        return ConnectionErrorState(message)


class InitialNetworkState(NetworkState):
    """
    The state of the network when the app starts.

    This is a transient state until
    the actual network state is obtained.
    """

    def found_previous_connection(self, app: Application) -> NetworkState:
        """
        A previously established connection was found.
        """
        state = get_network_state(app)
        if isinstance(state, (DisconnectedState, UnknownState)):
            app.interface_transition('no_active_connection_found')
        else:
            app.interface_transition('found_active_connection')
        return state

    def no_previous_connection_found(self, app: Application) -> NetworkState:
        """
        No previously established connection was found.

        This is probably the first time the app runs.
        """
        app.interface_transition('no_active_connection_found')
        return UnconnectedState()


class UnconnectedState(NetworkState):
    """
    There is no current connection active,
    nor is there a configured connection available.

    As soon as a server is chosen,
    that becomes the default
    and this state is no longer reached.
    """

    status_label = translated_property("Disconnected")


def connect_initial(app: Application) -> NetworkState:
    """
    Estabilish an initial connection to the server.
    """
    client = nm.get_client()
    assert app.current_network_uuid is not None
    nm.activate_connection(
        client,
        app.current_network_uuid,
        partial(on_any_update_callback, app),
    )
    return ConnectingState()


def connect(app: Application) -> NetworkState:
    """
    Estabilish a connection to the server.
    """
    client = nm.get_client()
    assert app.current_network_uuid is not None

    server = app.session_state.server
    if storage.get_connection_protocol(server) is Protocol.WIREGUARD:
        from eduvpn.interface.event import on_start_connection, on_chosen_profile
        oauth_session = storage.load_oauth_session(server)
        server_info = app.server_db.get_server_info(server)
        metadata = storage.get_current_metadata(server.oauth_login_url)
        assert metadata is not None
        profile_id = metadata[6]
        profile = server_info.get_profile(oauth_session, profile_id)
        if profile is None:
            # Profile was removed, redo profile choice.
            on_start_connection(
                app,
                server,
                oauth_session,
            )
        else:
            on_chosen_profile(
                app,
                server,
                oauth_session,
                profile,
            )
        return DisconnectedState()
    else:
        nm.activate_connection(
            client,
            app.current_network_uuid,
            partial(on_any_update_callback, app),
        )

    return ConnectingState()


def disconnect(app: Application, *, update_state=True) -> NetworkState:
    """
    Break the connection to the server.
    """
    client = nm.get_client()
    callback = None
    if update_state:
        callback = partial(on_any_update_callback, app)
    assert app.current_network_uuid is not None
    nm.deactivate_connection(
        client,
        app.current_network_uuid,
        callback,
    )
    return DisconnectedState()


def reconnect(app: Application) -> NetworkState:
    """
    Disconnect and then connect again.
    """
    disconnect(app, update_state=False)
    return ReconnectingState()


UNKNOWN_STATE_MAX_RETRIES = 5


def enter_unknown_state(app: Application) -> NetworkState:
    # Set the state temporarily to unknown but keep polling for updates,
    # since we don't always get notified by the update callback.
    @run_in_background_thread('poll-network-state')
    def determine_network_state_thread():
        counter = 0
        state = nm.get_connection_state()
        while state is nm.ConnectionState.UNKNOWN:
            if counter > UNKNOWN_STATE_MAX_RETRIES:
                # After a number of retries, assume we've disconnected
                # so the user can try to connect again.
                state = nm.ConnectionState.DISCONNECTED
                logger.debug(
                    "network state has been unknown for too long,"
                    " fall back to disconnected state"
                )
                break
            sleep(1)
            counter += 1
            if not isinstance(app.network_state, UnknownState):
                return
            state = nm.get_connection_state()
            logger.debug(f"polling network state: {state}")
        app.make_func_threadsafe(on_state_update_callback)(app, state)

    determine_network_state_thread()
    return UnknownState()


def on_state_update_callback(app: Application, state: nm.ConnectionState):
    """
    Callback for whenever a connection state changes.
    """
    if state is nm.ConnectionState.CONNECTING:
        app.network_transition('set_connecting')
    elif state is nm.ConnectionState.CONNECTED:
        app.network_transition('set_connected')
    elif state is nm.ConnectionState.DISCONNECTED:
        app.network_transition('set_disconnected')
    elif state is nm.ConnectionState.FAILED:
        app.network_transition('set_error')
    elif state is nm.ConnectionState.UNKNOWN:
        app.network_transition('set_unknown')
    else:
        raise ValueError(state)


def on_any_update_callback(app: Application):
    """
    Callback for whenever a connection state might have changed,
    where the new state is unknown.
    """
    state = nm.get_connection_state()
    on_state_update_callback(app, state)


def get_network_state(app: Application) -> NetworkState:
    """
    Get the network state that corresponds to the connection state.
    """
    state = nm.get_connection_state()
    return get_corresponding_state(app, state)


def get_corresponding_state(
    app: Application,
    state: nm.ConnectionState,
) -> NetworkState:
    """
    Get the network state that corresponds to the connection state.
    """
    if state is nm.ConnectionState.CONNECTING:
        return ConnectingState()
    elif state is nm.ConnectionState.CONNECTED:
        return ConnectedState()
    elif state is nm.ConnectionState.DISCONNECTED:
        return DisconnectedState()
    elif state is nm.ConnectionState.FAILED:
        return ConnectionErrorState(None)
    elif state is nm.ConnectionState.UNKNOWN:
        return enter_unknown_state(app)
    else:
        raise ValueError(state)


def check_network_state(app: Application):
    state = nm.get_connection_state()
    if state is not nm.ConnectionState.UNKNOWN:
        network_state = get_corresponding_state(app, state)
        if type(network_state) != type(app.network_state):  # NOQA
            # If the state remains different for one second,
            # update it in the app. This prevents
            # confusing transitions around connect/disconnect.
            sleep(1)
            if nm.get_connection_state() is state:
                app.make_func_threadsafe(on_state_update_callback)(app, state)


class ConnectingState(NetworkState):
    """
    The network is currently trying to connect to a server.
    """

    status_label = translated_property("Preparing to connect")
    status_image = StatusImage.CONNECTING

    def set_connecting(self, app: Application) -> 'NetworkState':
        return self

    def start_new_connection(self,
                             app: Application,
                             server: Server,
                             ) -> NetworkState:
        return reconnect(app)

    def disconnect(self, app: Application) -> NetworkState:
        """
        Abort connecting.
        """
        return disconnect(app)


class ConnectedState(NetworkState):
    """
    The network is currently connected to a server.
    """

    status_label = translated_property("Connection active")
    status_image = StatusImage.CONNECTED

    def set_connected(self, app: Application) -> 'NetworkState':
        return self

    def start_new_connection(self,
                             app: Application,
                             server: Server,
                             ) -> NetworkState:
        return reconnect(app)

    def disconnect(self, app: Application) -> NetworkState:
        return disconnect(app)


class DisconnectedState(NetworkState):
    """
    The network is not currently connected to a server,
    but a configured connection exists.
    """

    status_label = translated_property("Disconnected")
    status_image = StatusImage.NOT_CONNECTED

    def set_disconnected(self, app: Application) -> 'NetworkState':
        return self

    def reconnect(self, app: Application) -> NetworkState:
        return connect(app)


class ReconnectingState(NetworkState):
    """
    The network is currently disconnecting
    and will connect again when ready.
    """

    status_label = translated_property("Preparing to connect")
    status_image = StatusImage.CONNECTING

    def set_unknown(self, app: Application) -> NetworkState:
        return connect_initial(app)

    def set_disconnected(self, app: Application) -> NetworkState:
        return connect_initial(app)


class ConnectionErrorState(NetworkState):
    """
    The network could not connect because an error occured.
    """

    status_label = translated_property("Connection failed")
    status_image = StatusImage.NOT_CONNECTED

    def __init__(self, error: Optional[str]):
        self.error = error

    def reconnect(self, app: Application) -> NetworkState:
        return connect(app)


class UnknownState(NetworkState):
    """
    The network state could not be determined.
    """

    def reconnect(self, app: Application) -> NetworkState:
        return connect(app)



import logging
from typing import Tuple, List, Dict, Any
import json

import requests
from requests_oauthlib import OAuth2Session

from eduvpn.crypto import common_name_from_cert
from eduvpn.crypto import validate
from eduvpn.utils import add_retry_adapter
from eduvpn.settings import MAX_HTTP_RETRIES, MAX_HTTP_TIMEOUT

logger = logging.getLogger(__name__)


class InvalidProfile(Exception):
    def __init__(self, message):
        self.message = message


def request(uri: str, verify: bool = False) -> dict:
    """
    Do a request and check the signature using our public key verifier.
    """
    logger.info(f"Requesting {uri}")
    try:
        with add_retry_adapter(requests.Session(), MAX_HTTP_RETRIES) as session:
            response = session.get(uri, timeout=MAX_HTTP_TIMEOUT)
    except Exception as e:
        msg = f"Got exception {e} requesting {uri}"
        logger.debug(msg)
        raise
    if response.status_code != 200:
        msg = f"Got error code {response.status_code} requesting {uri}"
        logger.error(msg)
        raise IOError(msg)

    if verify:
        sig_uri = uri + '.minisig'
        logger.info(f"Retrieving signature {sig_uri}")
        sig_response = requests.get(sig_uri)
        if sig_response.status_code != 200:
            msg = f"Can't retrieve signature, requesting {sig_uri} gave error code {sig_response.status_code}"
            logger.error(msg)
            raise IOError(msg)

        logger.info(f"verifying signature of {uri}")
        signature = sig_response.content.decode('utf-8').split("\n")[1]
        _ = validate(content=response.content, signature=signature)
    return response.json()


def check_response(response: requests.Response):
    if response.status_code != 200:
        try:
            uri = response.history[0].url
        except IndexError:
            uri = '<empty>'
        msg = f"Got error code {response.status_code} requesting {uri}"
        logger.error(msg)
        raise IOError(msg)


def oauth_request(oauth: OAuth2Session, uri: str, method: str = 'get'):
    """
    Do an oauth request and check if there are no issues
    """
    call = getattr(oauth, method)
    response = call(uri)
    check_response(response)
    return response


def list_organisations(uri: str) -> List[Dict[str, Any]]:
    try:
        result = request(uri, verify=True)['organization_list']
    except Exception as e:
        logger.error(f"Got exception {e} requesting {uri} for organization_list")
        raise
    return result


def list_servers(uri: str) -> List[Dict[str, Any]]:
    try:
        result = request(uri, verify=True)['server_list']
    except Exception as e:
        logger.error(f"Got exception {e} requesting {uri} for server_list")
        raise
    return result


def get_full_info(base_uri: str) -> Dict[str, Any]:
    if not base_uri.endswith('/'):
        base_uri += '/'
    uri = base_uri + '.well-known/vpn-user-portal'
    return request(uri)['api']['http://eduvpn.org/api#2']


def get_info(base_uri: str):
    info = get_full_info(base_uri)
    api_base_uri = info['api_base_uri']
    token_endpoint = info['token_endpoint']
    auth_endpoint = info['authorization_endpoint']
    return api_base_uri, token_endpoint, auth_endpoint


def get_config(oauth: OAuth2Session, base_uri: str, profile_id: str) -> str:
    uri = base_uri + f'/profile_config?profile_id={profile_id}'
    text = oauth_request(oauth, uri).text
    try:
        data = json.loads(text)
    except json.decoder.JSONDecodeError:
        # On success, the response it *not* JSON but a valid ovpn file.
        pass
    else:
        # On errors, the server responds with JSON.
        if not data['profile_config']['ok']:
            logging.error(f"invalid profile_id: {profile_id} @ {base_uri}: {data}")
            raise InvalidProfile(data['profile_config']['error'])
        raise ValueError(data)
    return text


def list_profiles(oauth: OAuth2Session, api_base_uri: str):
    uri = api_base_uri + '/profile_list'
    return oauth_request(oauth, uri).json()['profile_list']['data']


def create_keypair(oauth: OAuth2Session, api_base_uri: str) -> Tuple[str, str]:
    uri = api_base_uri + '/create_keypair'
    keypair = oauth_request(oauth, uri, method='post').json()['create_keypair']['data']
    private_key = keypair['private_key']
    certificate = keypair['certificate']
    return private_key, certificate


def system_messages(oauth: OAuth2Session, api_base_uri: str):
    uri = api_base_uri + '/system_messages'
    return oauth_request(oauth, uri).json()['system_messages']['data']


def check_certificate(oauth: OAuth2Session, api_base_uri: str, certificate: str):
    common_name = common_name_from_cert(certificate.encode('ascii'))
    uri = api_base_uri + '/check_certificate?common_name=' + common_name
    return oauth_request(oauth, uri).json()['check_certificate']['data']['is_valid']



from pathlib import Path
from eduvpn.utils import get_prefix, get_config_dir

prefix = get_prefix()

CONFIG_PREFIX = (Path(get_config_dir()).expanduser() / "eduvpn").resolve()
CONFIG_DIR_MODE = 0o700
CONFIG_JSON_PREFIX = "2.0_"

DISCO_URI = 'https://disco.eduvpn.org/v2/'
ORGANISATION_URI = DISCO_URI + "organization_list.json"
SERVER_URI = DISCO_URI + "server_list.json"
HELP_URL = 'https://www.eduvpn.org'

CLIENT_ID = "org.eduvpn.app.linux"
SCOPE = ["config"]
CODE_CHALLENGE_METHOD = "S256"
LANGUAGE = 'nl'
COUNTRY = "nl-NL"

COUNTRY_MAP = Path(prefix + "/share/eduvpn/country_codes.json")
FLAG_PREFIX = prefix + "/share/eduvpn/images/flags/png/"
IMAGE_PREFIX = prefix + "/share/eduvpn/images/"
LC_IMAGE_PREFIX = prefix + "/share/letsconnect/images/"


# format: base64(<signature_algorithm> || <key_id> || <public_key>)
VERIFY_KEYS = [
    "RWRtBSX1alxyGX+Xn3LuZnWUT0w//B6EmTJvgaAxBMYzlQeI+jdrO6KF",
    "RWQKqtqvd0R7rUDp0rWzbtYPA3towPWcLDCl7eY9pBMMI/ohCmrS0WiM"
]
EDUVPN_ICON = prefix + "/share/icons/hicolor/128x128/apps/org.eduvpn.client.png"
EDUVPN_NAME = "eduVPN"
LETS_CONNECT_LOGO = LC_IMAGE_PREFIX + "letsconnect.png"
SERVER_ILLUSTRATION = LC_IMAGE_PREFIX + "server-illustration.png"
LETS_CONNECT_ICON = prefix + "/share/icons/hicolor/128x128/apps/org.letsconnect-vpn.client.png"
LETS_CONNECT_NAME = "Let's Connect!"

SESSION_PENDING_EXPIRY_MINUTES = 15
SESSION_PENDING_EXPIRY_FRACTION = 0.8

MAX_HTTP_TIMEOUT = 2
MAX_HTTP_RETRIES = 5



# eduvpngui - The GNU/Linux eduVPN GUI client
#
# Copyright: 2017-2020, The Commons Conservancy eduVPN Programme
# SPDX-License-Identifier: GPL-3.0+

from typing import Optional
import os
import webbrowser
import logging
from gettext import gettext as _, ngettext

import gi
gi.require_version('Gtk', '3.0')  # noqa: E402
gi.require_version('NM', '1.0')  # noqa: E402
from gi.repository import Gtk, GObject, GdkPixbuf

from ..settings import HELP_URL
from ..interface import state as interface_state
from .. import network as network_state
from ..server import CustomServer
from ..app import Application
from ..state_machine import (
    ENTER, EXIT, transition_callback, transition_edge_callback,
    transition_level_callback)
from ..session import Validity
from .. import session as session_state
from ..nm import nm_available, nm_managed
from ..utils import (
    get_prefix, run_in_main_gtk_thread, run_periodically, cancel_at_context_end)
from . import search
from .utils import show_ui_component, link_markup, show_error_dialog
from .stats import NetworkStats

logger = logging.getLogger(__name__)


UPDATE_EXIPRY_INTERVAL = 1.  # seconds

RENEWAL_ALLOW_FRACTION = .8


def get_validity_text(validity: Optional[Validity]) -> str:
    if validity is None:
        return _("Valid for <b>unknown</b>")
    if validity.is_expired:
        return _("This session has expired")
    delta = validity.remaining
    days = delta.days
    hours = delta.seconds // 3600
    if days == 0:
        if hours == 0:
            minutes = delta.seconds // 60
            return ngettext("Valid for <b>{0} minute</b>",
                            "Valid for <b>{0} minutes</b>", minutes).format(minutes)
        else:
            return ngettext("Valid for <b>{0} hour</b>",
                            "Valid for <b>{0} hours</b>", hours).format(hours)
    else:
        dstr = ngettext("Valid for <b>{0} day</b>",
                        "Valid for <b>{0} days</b>", days).format(days)
        hstr = ngettext(" and <b>{0} hour</b>",
                        " and <b>{0} hours</b>", hours).format(hours)
        return dstr + hstr


def get_template_path(filename: str) -> str:
    return os.path.join(get_prefix(), 'share/eduvpn/builder', filename)


class EduVpnGtkWindow(Gtk.ApplicationWindow):
    __gtype_name__ = "EduVpnGtkWindow"

    def __new__(cls, application: Application):
        builder = Gtk.Builder()
        builder.add_from_file(get_template_path('mainwindow.ui'))  # type: ignore
        window = builder.get_object('eduvpn')  # type: ignore
        window.setup(builder, application)  # type: ignore
        window.set_application(application)  # type: ignore
        return window

    def setup(self, builder, application: Application):
        self.app = application.app  # type: ignore
        handlers = {
            "on_configure_settings": self.on_configure_settings,
            "on_get_help": self.on_get_help,
            "on_go_back": self.on_go_back,
            "on_add_other_server": self.on_add_other_server,
            "on_add_custom_server": self.on_add_custom_server,
            "on_cancel_oauth_setup": self.on_cancel_oauth_setup,
            "on_select_server": self.on_select_server,
            "on_search_changed": self.on_search_changed,
            "on_search_activate": self.on_search_activate,
            "on_switch_connection_state": self.on_switch_connection_state,
            "on_toggle_connection_info": self.on_toggle_connection_info,
            "on_profile_selection_changed": self.on_profile_selection_changed,
            "on_location_selection_changed": self.on_location_selection_changed,
            "on_acknowledge_error": self.on_acknowledge_error,
            "on_renew_session_clicked": self.on_renew_session_clicked,
            "on_config_force_tcp": self.on_config_force_tcp,
            "on_close_window": self.on_close_window,
        }
        builder.connect_signals(handlers)

        self.app_logo = builder.get_object('appLogo')

        self.page_stack = builder.get_object('pageStack')
        self.settings_button = builder.get_object('settingsButton')
        self.back_button_container = builder.get_object('backButtonEventBox')

        self.server_list_container = builder.get_object('serverListContainer')

        self.institute_list_header = builder.get_object('instituteAccessHeader')
        self.secure_internet_list_header = builder.get_object('secureInternetHeader')
        self.other_server_list_header = builder.get_object('otherServersHeader')

        self.institute_list = builder.get_object('instituteTreeView')
        self.secure_internet_list = builder.get_object('secureInternetTreeView')
        self.other_server_list = builder.get_object('otherServersTreeView')

        self.choose_profile_page = builder.get_object('chooseProfilePage')
        self.choose_location_page = builder.get_object('chooseLocationPage')
        self.location_list = builder.get_object('locationTreeView')
        self.profile_list = builder.get_object('profileTreeView')

        self.find_server_page = builder.get_object('findServerPage')
        self.find_server_search_form = builder.get_object('findServerSearchForm')
        self.find_server_search_input = builder.get_object('findServerSearchInput')
        self.find_server_image = builder.get_object('findServerImage')
        self.find_server_label = builder.get_object('findServerLabel')

        self.add_custom_server_button_container = builder.get_object('addCustomServerRow')
        self.add_other_server_button_container = builder.get_object('addOtherServerRow')

        self.connection_page = builder.get_object('connectionPage')
        self.connection_status_image = builder.get_object('connectionStatusImage')
        self.connection_status_label = builder.get_object('connectionStatusLabel')
        self.connection_session_label = builder.get_object('connectionSessionLabel')
        self.connection_switch = builder.get_object('connectionSwitch')
        self.connection_info_expander = builder.get_object('connectionInfoExpander')
        self.connection_info_downloaded = builder.get_object('connectionInfoDownloadedText')
        self.connection_info_uploaded = builder.get_object('connectionInfoUploadedText')
        self.connection_info_ipv4address = builder.get_object('connectionInfoIpv4AddressText')
        self.connection_info_ipv6address = builder.get_object('connectionInfoIpv6AddressText')
        self.connection_info_thread_cancel = None
        self.connection_info_stats = None

        self.server_image = builder.get_object('serverImage')
        self.server_label = builder.get_object('serverLabel')
        self.server_support_label = builder.get_object('supportLabel')

        self.renew_session_button = builder.get_object('renewSessionButton')

        self.oauth_page = builder.get_object('openBrowserPage')
        self.oauth_cancel_button = builder.get_object('cancelBrowserButton')

        self.settings_page = builder.get_object('settingsPage')
        self.setting_config_force_tcp = builder.get_object('settingConfigForceTCP')

        self.loading_page = builder.get_object('loadingPage')
        self.loading_title = builder.get_object('loadingTitle')
        self.loading_message = builder.get_object('loadingMessage')

        self.error_page = builder.get_object('errorPage')
        self.error_text = builder.get_object('errorText')
        self.error_acknowledge_button = builder.get_object('errorAcknowledgeButton')

        self.set_title(self.app.variant.name)  # type: ignore
        self.set_icon_from_file(self.app.variant.icon)  # type: ignore
        if self.app.variant.logo:
            self.app_logo.set_from_file(self.app.variant.logo)
        if self.app.variant.server_image:
            self.find_server_image.set_from_file(self.app.variant.server_image)
        if not self.app.variant.use_predefined_servers:
            self.find_server_label.set_text(_("Server address"))
            self.find_server_search_input.set_placeholder_text(_("Enter the server address"))

        # Track the currently shown page so we can return to it
        # when the settings page is closed.
        self.current_shown_page = None

        # We track the switch state so we can distinguish
        # the switch being set by the ui from the user toggling it.
        self.connection_switch_state: Optional[bool] = None

    def initialize(self):
        self.app.connect_state_transition_callbacks(self)

        if not nm_available():
            show_error_dialog(
                self,
                name=_("Error"),
                title=_("NetworkManager not available"),
                message=_("The application will not be able to configure the network. Please install and set up NetworkManager."))
        elif not nm_managed():
            show_error_dialog(
                self,
                name=_("Error"),
                title=_("NetworkManager not managing device"),
                message=_("The application will not be able to configure the network. NetworkManager is installed but no device of the primary connection is currently managed by it."))

    # ui functions

    def show_back_button(self, show: bool):
        show_ui_component(self.back_button_container, show)

    def set_search_text(self, text: str):
        self.find_server_search_input.set_text(text)

    def show_loading_page(self, title: str, message: str):
        self.show_page(self.loading_page)
        self.loading_title.set_text(title)
        self.loading_message.set_text(message)

    def hide_loading_page(self):
        self.hide_page(self.loading_page)

    def set_connection_switch_state(self, state: bool):
        self.connection_switch_state = state
        self.connection_switch.set_state(state)

    def show_page(self, page):
        """
        Show a collection of pages.
        """
        self.page_stack.set_visible_child(page)
        self.current_shown_page = page

    def hide_page(self, page):
        """
        Show a collection of pages.
        """
        self.current_shown_page = None

    def is_on_settings_page(self) -> bool:
        return self.page_stack.get_visible_child() is self.settings_page

    def enter_settings_page(self):
        assert not self.is_on_settings_page()
        self.setting_config_force_tcp.set_state(self.app.config.force_tcp)
        self.page_stack.set_visible_child(self.settings_page)
        self.show_back_button(True)

    def leave_settings_page(self):
        assert self.is_on_settings_page()
        self.page_stack.set_visible_child(self.current_shown_page)
        self.show_back_button(self.app.interface_state.has_transition('go_back'))

    # network state transition callbacks

    @transition_callback(network_state.NetworkState)
    def default_network_transition_callback(self, old_state, new_state):
        if isinstance(self.app.interface_state, interface_state.ConnectionStatus):
            self.update_connection_status()

    def update_connection_server(self):
        if isinstance(self.app.session_state,
                      (session_state.InitialSessionState,
                       session_state.NoSessionState)):
            self.server_label.set_text('')
            self.server_image.hide()
            self.server_support_label.hide()
            return
        else:
            server = self.app.session_state.server

        self.server_label.set_text(str(server))

        server_image_path = getattr(server, 'image_path', None)
        if server_image_path:
            self.server_image.set_from_file(server_image_path)
            self.server_image.show()
        else:
            self.server_image.hide()

        if getattr(server, 'support_contact', []):
            support_text = _("Support:") + "\n" + "\n".join(map(link_markup, server.support_contact))
            self.server_support_label.set_markup(support_text)
            self.server_support_label.show()
        else:
            self.server_support_label.hide()

    def update_connection_validity(self):
        if isinstance(self.app.session_state,
                      (session_state.InitialSessionState,
                       session_state.NoSessionState)):
            self.connection_session_label.hide()
        else:
            expiry_text = get_validity_text(self.app.session_state.validity)
            self.connection_session_label.show()
            self.connection_session_label.set_markup(expiry_text)

    def update_connection_status(self):
        self.connection_status_label.set_text(self.app.network_state.status_label)
        self.connection_status_image.set_from_file(self.app.network_state.status_image.path)

        self.update_connection_validity()

        assert not (self.app.network_state.has_transition('reconnect') and self.app.network_state.has_transition('disconnect'))
        if self.app.network_state.has_transition('reconnect'):
            if self.app.session_state.is_active:
                self.connection_switch.show()
                self.set_connection_switch_state(False)
            else:
                self.connection_switch.hide()
        elif self.app.network_state.has_transition('disconnect'):
            self.connection_switch.show()
            self.set_connection_switch_state(True)
        else:
            self.connection_switch.hide()

        if self.app.session_state.has_transition('renew'):
            if isinstance(self.app.session_state, session_state.SessionExpiredState) and not isinstance(self.app.network_state, network_state.DisconnectedState):
                # The user needs to disconnect first.
                self.renew_session_button.hide()
            else:
                self.renew_session_button.show()
        else:
            self.renew_session_button.hide()

    # session state transition callbacks

    @transition_callback(session_state.SessionState)
    def default_session_transition_callback(self, old_state, new_state):
        if isinstance(self.app.interface_state, interface_state.ConnectionStatus):
            self.update_connection_status()

    # interface state transition callbacks

    @transition_callback(interface_state.InterfaceState)
    def default_interface_transition_callback(self, old_state, new_state):
        # Only show the 'go back' button if
        # the corresponding transition is available.
        self.show_back_button(new_state.has_transition('go_back'))

    @transition_edge_callback(ENTER, interface_state.configure_server_states)
    def enter_search(self, old_state, new_state):
        if not isinstance(old_state, interface_state.configure_server_states):
            self.find_server_search_input.grab_focus()
            search.show_result_components(self, True)
            search.show_search_components(self, True)
            search.init_server_search(self)
            search.connect_selection_handlers(self, self.on_select_server)

    @transition_edge_callback(EXIT, interface_state.configure_server_states)
    def exit_search(self, old_state, new_state):
        if not isinstance(new_state, interface_state.configure_server_states):
            search.show_result_components(self, False)
            search.show_search_components(self, False)
            search.exit_server_search(self)
            search.disconnect_selection_handlers(self, self.on_select_server)
            self.set_search_text('')

    @transition_edge_callback(
        ENTER, interface_state.PendingConfigurePredefinedServer)
    def enter_PendingConfigurePredefinedServer(self, old_state, new_state):
        search.update_results(self, [])
        if not isinstance(old_state, interface_state.configure_server_states):
            self.set_search_text(new_state.search_query)

    @transition_edge_callback(ENTER, interface_state.ConfigurePredefinedServer)
    def enter_ConfigurePredefinedServer(self, old_state, new_state):
        search.update_results(self, new_state.results)
        if not isinstance(old_state, interface_state.configure_server_states):
            self.set_search_text(new_state.search_query)

    @transition_edge_callback(ENTER, interface_state.ConfigureCustomServer)
    def enter_ConfigureCustomServer(self, old_state, new_state):
        if self.app.variant.use_predefined_servers:
            search.update_results(self, [CustomServer(new_state.address)])
            if not isinstance(old_state, interface_state.configure_server_states):
                self.set_search_text(new_state.address)
        else:
            entered_address = len(new_state.address) > 0
            show_ui_component(self.add_custom_server_button_container, entered_address)

    @transition_edge_callback(EXIT, interface_state.ConfigureCustomServer)
    def exit_ConfigureCustomServer(self, old_state, new_state):
        if not self.app.variant.use_predefined_servers:
            self.add_custom_server_button_container.hide()

    @transition_edge_callback(ENTER, interface_state.MainState)
    def enter_MainState(self, old_state, new_state):
        search.show_result_components(self, True)
        self.add_other_server_button_container.show()
        search.update_results(self, new_state.servers)
        search.init_server_search(self)
        search.connect_selection_handlers(self, self.on_select_server)

    @transition_edge_callback(EXIT, interface_state.MainState)
    def exit_MainState(self, old_state, new_state):
        search.show_result_components(self, False)
        self.add_other_server_button_container.hide()
        search.exit_server_search(self)
        search.disconnect_selection_handlers(self, self.on_select_server)

    @transition_edge_callback(ENTER, interface_state.OAuthSetupPending)
    @transition_edge_callback(ENTER, interface_state.OAuthSetup)
    def enter_oauth_setup(self, old_state, new_state):
        self.show_page(self.oauth_page)
        in_setup_state = isinstance(new_state, interface_state.OAuthSetup)
        show_ui_component(self.oauth_cancel_button, in_setup_state)

    @transition_edge_callback(EXIT, interface_state.OAuthSetupPending)
    @transition_edge_callback(EXIT, interface_state.OAuthSetup)
    def exit_oauth_setup(self, old_state, new_state):
        self.hide_page(self.oauth_page)
        self.oauth_cancel_button.hide()

    @transition_edge_callback(ENTER, interface_state.OAuthRefreshToken)
    def enter_OAuthRefreshToken(self, old_state, new_state):
        self.show_loading_page(
            _("Finishing Authorization"),
            _("The authorization token is being finished."),
        )

    @transition_edge_callback(EXIT, interface_state.OAuthRefreshToken)
    def exit_OAuthRefreshToken(self, old_state, new_state):
        self.hide_loading_page()

    @transition_edge_callback(ENTER, interface_state.LoadingServerInformation)
    def enter_LoadingServerInformation(self, old_state, new_state):
        self.show_loading_page(
            _("Loading"),
            _("The server details are being loaded."),
        )

    @transition_edge_callback(EXIT, interface_state.LoadingServerInformation)
    def exit_LoadingServerInformation(self, old_state, new_state):
        self.hide_loading_page()

    @transition_edge_callback(ENTER, interface_state.ChooseProfile)
    def enter_ChooseProfile(self, old_state, new_state):
        self.show_page(self.choose_profile_page)
        self.profile_list.show()

        profile_tree_view = self.profile_list
        profiles_list_model = Gtk.ListStore(GObject.TYPE_STRING, GObject.TYPE_PYOBJECT)

        if len(profile_tree_view.get_columns()) == 0:
            # Only initialize this tree view once.
            text_cell = Gtk.CellRendererText()
            text_cell.set_property("size-points", 14)

            column = Gtk.TreeViewColumn(None, text_cell, text=0)
            profile_tree_view.append_column(column)

        profile_tree_view.set_model(profiles_list_model)
        profiles_list_model.clear()
        for profile in new_state.profiles:
            profiles_list_model.append([str(profile), profile])

    @transition_edge_callback(EXIT, interface_state.ChooseProfile)
    def exit_ChooseProfile(self, old_state, new_state):
        self.hide_page(self.choose_profile_page)
        self.profile_list.hide()

    @transition_edge_callback(ENTER, interface_state.ChooseSecureInternetLocation)
    def enter_ChooseSecureInternetLocation(self, old_state, new_state):
        self.show_page(self.choose_location_page)
        self.location_list.show()

        location_tree_view = self.location_list
        location_list_model = Gtk.ListStore(GObject.TYPE_STRING, GdkPixbuf.Pixbuf, GObject.TYPE_PYOBJECT)

        if len(location_tree_view.get_columns()) == 0:
            # Only initialize this tree view once.
            text_cell = Gtk.CellRendererText()
            text_cell.set_property("size-points", 14)

            renderer_pixbuf = Gtk.CellRendererPixbuf()
            column = Gtk.TreeViewColumn("Image", renderer_pixbuf, pixbuf=1)
            location_tree_view.append_column(column)

            column = Gtk.TreeViewColumn(None, text_cell, text=0)
            location_tree_view.append_column(column)

            location_tree_view.set_model(location_list_model)

        location_list_model.clear()
        for location in new_state.locations:
            if location.flag_path is None:
                logger.warning(f"No flag found for country code {location.country_code}")
                flag = None
            else:
                flag = GdkPixbuf.Pixbuf.new_from_file(location.flag_path)
            location_list_model.append([location.country_name, flag, location])

    @transition_edge_callback(EXIT, interface_state.ChooseSecureInternetLocation)
    def exit_ChooseSecureInternetLocation(self, old_state, new_state):
        self.hide_page(self.choose_location_page)
        self.location_list.hide()

    @transition_edge_callback(ENTER, interface_state.ConfiguringConnection)
    def enter_ConfiguringConnection(self, old_state, new_state):
        self.show_loading_page(
            _("Configuring"),
            _("Your connection is being configured."),
        )

    @transition_edge_callback(EXIT, interface_state.ConfiguringConnection)
    def exit_ConfiguringConnection(self, old_state, new_state):
        self.hide_loading_page()

    @transition_edge_callback(ENTER, network_state.ConnectedState)
    def enter_ConnectedState(self, old_state, new_state):
        is_expanded = self.connection_info_expander.get_expanded()
        if is_expanded:
            self.start_connection_info()

    @transition_edge_callback(ENTER, interface_state.ConnectionStatus)
    def enter_ConnectionStatus(self, old_state, new_state):
        self.show_page(self.connection_page)
        self.update_connection_server()
        self.update_connection_status()

    @transition_edge_callback(EXIT, interface_state.ConnectionStatus)
    def exit_ConnectionStatus(self, old_state, new_state):
        self.hide_page(self.connection_page)
        self.pause_connection_info()

    @transition_level_callback(interface_state.ConnectionStatus)
    def context_ConnectionStatus(self, state):
        return cancel_at_context_end(run_periodically(
            run_in_main_gtk_thread(self.update_connection_validity),
            UPDATE_EXIPRY_INTERVAL,
            'update-validity',
        ))

    @transition_edge_callback(ENTER, network_state.DisconnectedState)
    def enter_DisconnectedState(self, old_state, new_state):
        self.stop_connection_info()

    @transition_edge_callback(ENTER, interface_state.ErrorState)
    def enter_ErrorState(self, old_state, new_state):
        self.show_page(self.error_page)
        self.error_text.set_text(new_state.message)
        has_next_transition = new_state.next_transition is not None
        show_ui_component(self.error_acknowledge_button, has_next_transition)

    @transition_edge_callback(EXIT, interface_state.ErrorState)
    def exit_ErrorState(self, old_state, new_state):
        self.hide_page(self.error_page)

    # ui callbacks

    def on_configure_settings(self, widget, event):
        logger.debug("clicked on configure settings")
        if self.is_on_settings_page():
            self.leave_settings_page()
        else:
            self.enter_settings_page()

    def on_get_help(self, widget, event):
        logger.debug("clicked on get help")
        webbrowser.open(HELP_URL)

    def on_go_back(self, widget, event):
        logger.debug("clicked on go back")
        if self.is_on_settings_page():
            self.leave_settings_page()
        else:
            self.app.interface_transition('go_back')

    def on_add_other_server(self, button) -> None:
        logger.debug("clicked on add other server")
        self.app.interface_transition('configure_new_server')

    def on_add_custom_server(self, button) -> None:
        logger.debug("clicked on add custom server")
        server = CustomServer(self.app.interface_state.address)
        self.app.interface_transition('connect_to_server', server)

    def on_select_server(self, selection):
        logger.debug("selected server search result")
        (model, tree_iter) = selection.get_selected()
        selection.unselect_all()
        if tree_iter is None:
            logger.info("selection empty")
        else:
            row = model[tree_iter]
            server = row[1]
            logger.debug(f"selected server: {server!r}")
            self.app.interface_transition('connect_to_server', server)

    def on_cancel_oauth_setup(self, _):
        logger.debug("clicked on cancel oauth setup")
        self.app.interface_transition('oauth_setup_cancel')

    def on_search_changed(self, _=None):
        query = self.find_server_search_input.get_text()
        logger.debug(f"entered server search query: {query}")
        if self.app.variant.use_predefined_servers and query.count('.') < 2:
            self.app.interface_transition(
                'enter_search_query', search_query=query)
        else:
            # Anything with two periods is interpreted
            # as a custom server address.
            self.app.interface_transition(
                'enter_custom_address', address=query)

    def on_search_activate(self, _=None):
        logger.debug("activated server search")
        # TODO

    def on_switch_connection_state(self, switch, state):
        logger.debug("clicked on switch connection state")
        if state is not self.connection_switch_state:
            self.connection_switch_state = state
            # The user has toggled the connection switch,
            # as opposed to the ui itself setting it.
            if state:
                self.app.interface_transition('activate_connection')
            else:
                self.app.interface_transition('deactivate_connection')
        return True

    def pause_connection_info(self):
        if self.connection_info_thread_cancel:
            self.connection_info_thread_cancel()
            self.connection_info_thread_cancel = None

    def stop_connection_info(self):
        # Pause the thread
        self.pause_connection_info()

        # Further cleanup
        if self.connection_info_stats:
            self.connection_info_stats.cleanup()
            self.connection_info_stats = None

    def start_connection_info(self):
        if not self.app.network_state.has_transition('disconnect'):
            logger.info("Connection Info: VPN is not active")
            return

        def update_connection_info_callback():
            # Do nothing if we have no stats object
            if not self.connection_info_stats:
                return
            download = self.connection_info_stats.download
            upload = self.connection_info_stats.upload
            ipv4 = self.connection_info_stats.ipv4
            ipv6 = self.connection_info_stats.ipv6
            self.connection_info_downloaded.set_text(download)
            self.connection_info_uploaded.set_text(upload)
            self.connection_info_ipv4address.set_text(ipv4)
            self.connection_info_ipv6address.set_text(ipv6)

        if not self.connection_info_stats:
            self.connection_info_stats = NetworkStats()

        if not self.connection_info_thread_cancel:
            # Run every second in the background
            self.connection_info_thread_cancel = run_periodically(
                update_connection_info_callback, 1
            )

    def on_toggle_connection_info(self, _):
        logger.debug("clicked on connection info")
        was_expanded = self.connection_info_expander.get_expanded()

        if not was_expanded:
            self.start_connection_info()
        else:
            self.pause_connection_info()

    def on_profile_selection_changed(self, selection):
        logger.debug("selected profile")
        (model, tree_iter) = selection.get_selected()
        selection.unselect_all()
        if tree_iter is None:
            logger.debug("selection empty")
        else:
            row = model[tree_iter]
            profile = row[1]
            logger.debug(f"selected profile: {profile!r}")
            self.app.interface_transition('select_profile', profile)

    def on_location_selection_changed(self, selection):
        logger.debug("selected location")
        (model, tree_iter) = selection.get_selected()
        selection.unselect_all()
        if tree_iter is None:
            logger.debug("selection empty")
        else:
            row = model[tree_iter]
            location = row[2]
            logger.debug(f"selected location: {location!r}")
            self.app.interface_transition('select_secure_internet_location', location)

    def on_acknowledge_error(self, event):
        logger.debug("clicked on acknowledge error")
        self.app.interface_transition('acknowledge_error')

    def on_renew_session_clicked(self, event):
        logger.debug("clicked on renew session")
        self.app.session_transition('renew')

    def on_config_force_tcp(self, switch, state: bool):
        logger.debug("clicked on setting: 'force tcp'")
        self.app.config.force_tcp = state

    def on_close_window(self, window, event):
        logger.debug("clicked on close window")
        self.hide()
        self.get_application().on_window_closed()
        return True

    def on_reopen_window(self):
        self.app.interface_transition('restart')
        self.show()
        self.present()



from logging import getLogger
from argparse import Namespace
from itertools import chain
from pathlib import Path
from sys import exit
from typing import List, Dict, Optional, Tuple, Any

import nacl.exceptions

from eduvpn.i18n import extract_translation
from eduvpn.nm import nm_available, save_connection_with_mainloop
from eduvpn.remote import list_servers, list_organisations
from eduvpn.settings import SERVER_URI, ORGANISATION_URI
from eduvpn.storage import write_config

_logger = getLogger()

ServerListType = List[Dict[str, Any]]


def fetch_servers_orgs() -> Tuple[ServerListType, ServerListType]:
    servers = list_servers(SERVER_URI)
    orgs = list_organisations(ORGANISATION_URI)
    return servers, orgs


def input_int(max_: int):
    """
    Request the user to enter a number.
    """
    while True:
        choice = input("\n> ")
        if choice.isdigit() and int(choice) < max_:
            break
        else:
            print("error: invalid choice")
    return int(choice)


def provider_choice(institutes: List[dict], orgs: List[dict]) -> Tuple[str, str, Optional[str], bool]:
    """
    Ask the user to make a choice from a list of institute and secure internet providers.

    returns:
        url, display_name, contact, bool. Bool indicates if it is secure_internet or not.
    """
    print("\nPlease choose server:\n")
    print("Institute access:")
    for i, row in enumerate(institutes):
        print(f"[{i}] {extract_translation(row['display_name'])}")

    print("Secure internet: \n")
    for i, row in enumerate(orgs, start=len(institutes)):
        print(f"[{i}] {extract_translation(row['display_name'])}")

    choice = input_int(max_=len(institutes) + len(orgs))

    if choice < len(institutes):
        institute = institutes[choice]
        return institute['base_url'], extract_translation(institute['display_name']), institute[
            'support_contact'], False
    else:
        org = orgs[choice - len(institutes)]
        return org['secure_internet_home'], extract_translation(org['display_name']), None, True


def menu(
        institutes: List[dict],
        orgs: List[dict],
        search_term: Optional[str] = None
) -> Tuple[str, str, Optional[str], bool]:
    """
    returns:
        url, bool. Bool indicates if it is secure_internet or not.
    """
    # todo: add initial search filtering
    return provider_choice(institutes, orgs)


def profile_choice(profiles: List[Dict]) -> str:
    """
    If multiple profiles are available, present user with choice which profile.
    """
    if len(profiles) > 1:
        print("\nplease choose a profile:\n")
        for i, profile in enumerate(profiles):
            print(f" * [{i}] {profile['display_name']}")
        choice = input_int(max_=len(profiles))
        return profiles[int(choice)]['profile_id']
    else:
        return profiles[0]['profile_id']


def write_to_nm_choice() -> bool:
    """
    When Network Manager is available, asks user to add VPN to Network Manager
    """
    print("\nWhat would you like to do with your VPN configuration:\n")
    print("* [0] Write .ovpn file to current directory")
    print("* [1] Add VPN configuration to Network Manager")
    return bool(input_int(max_=2))


def secure_internet_choice(secure_internets: List[dict]) -> Optional[Tuple[str, str]]:
    print("Do you want to select a secure internet location? If not we use the default.")
    while True:
        choice = input("\n[N/y] > ").strip().lower()
        if choice == 'n' or not choice:
            return None
        elif choice == 'y':
            print("\nplease choose a secure internet server:\n")
            for i, profile in enumerate(secure_internets):
                print(f" * [{i}] {extract_translation(profile['country_code'])}")
            choice = input_int(max_=len(secure_internets))
            base_url = secure_internets[int(choice)]['base_url']
            country_code = secure_internets[int(choice)]['country_code']
            return base_url, country_code
        else:
            print("error: invalid choice, please enter y, n or just leave empty")


def search(args: Namespace):
    search_term = args.match
    servers, orgs = fetch_servers_orgs()
    institute_matches, org_matches = match_term(servers, orgs, search_term)
    print(f"Your search term '{search_term}'  matched with the following institutes/organisations:\n")

    if len(institute_matches):
        print("Institute access:")
        for i, row in institute_matches:
            print(f"[{i}] {extract_translation(row['display_name'])}")

    if len(org_matches):
        print("\nSecure internet: \n")
        for i, row in org_matches:
            print(f"[{i}] {extract_translation(row['display_name'])}")


def configure(args: Namespace) -> Tuple[str, str, Optional[str], Optional[ServerListType]]:
    search_term = args.match
    servers, orgs = fetch_servers_orgs()
    secure_internets = [s for s in servers if s['server_type'] == 'secure_internet']
    institute_matches, org_matches = match_term(servers, orgs, search_term, exact=True)

    if isinstance(search_term, str) and search_term.lower().startswith('https://'):
        return search_term, search_term, None, None
    else:
        if len(institute_matches) == 0 and len(org_matches) == 0:
            print(f"The filter '{search_term}' had no matches")
            exit(1)
        elif len(institute_matches) == 1 and len(org_matches) == 0:
            index, institute = institute_matches[0]
            print(f"filter '{search_term}' matched with institute '{institute['display_name']}'")
            return institute['base_url'], extract_translation(institute['display_name']), institute[
                'support_contact'], None
        elif len(institute_matches) == 0 and len(org_matches) == 1:
            index, org = org_matches[0]
            print(f"filter '{search_term}' matched with organisation '{org['display_name']}'")
            return org['secure_internet_home'], extract_translation(org['display_name']), None, secure_internets
        else:
            matches = [i[1]['display_name'] for i in chain(institute_matches, org_matches)]
            print(
                f"filter '{search_term}' matched with {len(matches)} institutes and organisations, please be more specific.")
            print("Matches:")
            for m in matches:
                print(f" - {extract_translation(m)}")
            exit(1)


def interactive(args: Namespace) -> Tuple[str, str, Optional[str], Optional[ServerListType]]:
    """
    returns:
        auth_url, display_name, support_contact, secure_internets
    """
    search_term = args.match

    if isinstance(search_term, str) and search_term.lower().startswith('https://'):
        return search_term, search_term, None, None

    try:
        servers = list_servers(SERVER_URI)
    except nacl.exceptions.BadSignatureError:
        print(f"Received a bad signature from server {SERVER_URI}")
        exit(1)
    secure_internets = [s for s in servers if s['server_type'] == 'secure_internet']
    institute_access = [s for s in servers if s['server_type'] == 'institute_access']
    orgs = list_organisations(ORGANISATION_URI)
    choice = menu(institutes=institute_access, orgs=orgs, search_term=search_term)

    if not choice:
        exit(1)

    auth_url, display_name, support_contact, secure_internet = choice
    if not secure_internet:
        return auth_url, display_name, support_contact, None

    return auth_url, display_name, support_contact, secure_internets


def match_term(
        servers: ServerListType,
        orgs: ServerListType,
        search_term: Optional[str], exact=False
) -> Tuple[List[Tuple[int, Dict[str, Any]]], List[Tuple[int, Dict[str, Any]]]]:
    """
    Search the list of institutes and organisations for a string match.

    returns:
        None or (type ('base_url' or 'secure_internet_home'), url)
    """
    institute_access = [s for s in servers if s['server_type'] == 'institute_access']

    if not search_term:
        return list(enumerate(institute_access)), list(enumerate(orgs, len(institute_access)))

    institute_matches: List[Tuple[int, Dict[str, Any]]] = []
    for x, i in enumerate(institute_access):
        if not exact:
            if search_term.lower() in extract_translation(i['display_name']).lower():
                institute_matches.append((x, i))
        else:
            if search_term.lower() == extract_translation(i['display_name']).lower():
                institute_matches.append((x, i))

    org_matches: List[Tuple[int, Dict[str, Any]]] = []
    for x, i in enumerate(orgs, len(institute_access)):
        if not exact:
            if search_term.lower() in extract_translation(i['display_name']).lower() \
                    or 'keyword_list' in i and search_term in i['keyword_list']:
                org_matches.append((x, i))
        else:
            if search_term.lower() == extract_translation(i['display_name']).lower():
                org_matches.append((x, i))
    return institute_matches, org_matches


def store_configuration(config, private_key, certificate, interactive=False):
    target = Path('eduVPN.ovpn').resolve()
    if interactive and nm_available():
        if write_to_nm_choice():
            save_connection_with_mainloop(config, private_key, certificate)
        else:
            write_config(config, private_key, certificate, target)
    else:
        if nm_available():
            save_connection_with_mainloop(config, private_key, certificate)
        else:
            write_config(config, private_key, certificate, target)



from io import StringIO
from typing import Optional, Callable, Iterable, List
from gettext import gettext as _


class Item:
    def to_string(self) -> str:
        raise NotImplementedError

    def write(self, file):
        line = self.to_string()
        file.write(f'{line}\n')

    def __eq__(self, other):
        return self.__class__ is other.__class__ and self.__dict__ == other.__dict__

    def __repr__(self):
        fields = ','.join(f' {k}={v!r}' for k, v in self.__dict__.items())
        return f'<{self.__class__.__name__}{fields}>'


class Field(Item):
    def __init__(self, name: str, arguments: List[str]):
        self.name = name
        self.arguments = arguments

    def to_string(self):
        return f'{self.name} {" ".join(self.arguments)}'


class Section(Item):
    def __init__(self, tag: str, content: List[str]):
        self.tag = tag
        self.content = content

    def to_string(self):
        lines = []
        lines.append(f'<{self.tag}>')
        lines.extend(self.content)
        lines.append(f'</{self.tag}>')
        return '\n'.join(lines)


class Comment(Item):
    def __init__(self, content: str):
        self.content = content

    def to_string(self):
        return f'#{self.content}'


class Empty(Item):
    def to_string(self):
        return ''


class InvalidOVPN(Exception):
    pass


def parse_ovpn(lines: Iterable[str]) -> Iterable[Item]:
    current_section = None
    for lineno, line in enumerate(lines):
        if current_section is not None:
            if line.strip() == f'</{current_section.tag}>':
                yield current_section
                current_section = None
            else:
                current_section.content.append(line)
        elif line.startswith('#'):
            yield Comment(line.rstrip()[1:])
        elif line.startswith('<'):
            line = line.rstrip()
            assert line.endswith('>')
            section = Section(line[1:-1], [])
            current_section = section
        elif line.rstrip() == '':
            yield Empty()
        else:
            field_name, *arguments = line.rstrip().split()
            yield Field(field_name, arguments)
    assert current_section is None


class Ovpn:
    def __init__(self, content: List[Item]):
        self.content = content

    @classmethod
    def parse(cls, content: str) -> 'Ovpn':
        return cls(list(parse_ovpn(content.splitlines())))

    def write(self, file):
        for item in self.content:
            item.write(file)

    def to_string(self) -> str:
        file = StringIO()
        self.write(file)
        return file.getvalue()

    def contains_field(self, field_name: str) -> bool:
        for item in self.content:
            if isinstance(item, Field):
                if item.name == field_name:
                    return True
        return False

    def replace_fields(self, replace: Callable[[Field], Optional[Item]]):
        new_content = []
        for item in self.content:
            if isinstance(item, Field):
                replacement = replace(item)
                if replacement is not None:
                    new_content.append(replacement)
            else:
                new_content.append(item)
        self.content = new_content

    def force_tcp(self):
        def replacement(field):
            if field.name == 'remote' and field.arguments[2] == 'udp':
                return Comment(f' omitted to force tcp: {field.to_string()}')
            else:
                return field

        self.replace_fields(replacement)
        if not self.contains_field('remote'):
            raise InvalidOVPN(_(
                "This connection does not provide TCP connections. "
                "Disable the setting 'force-tcp' to use it."
            ))



import os
from glob import glob

from setuptools import setup, find_packages

__version__ = "3.0.0"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()


install_requires = [
    'requests',
    'requests_oauthlib',
    'cryptography',
    'pynacl',
    'wheel',
]

tests_require = [
    'pytest',
    'pycodestyle',
]

mypy_require = [
    'mypy',
    'PyGObject-stubs',
    'types-requests',
    'types-setuptools',
]

gui_require = [
    'dbus-python',
    'pygobject',
]

extras_require = {
    'gui': gui_require,
    'test': tests_require,
    'mypy': mypy_require,
}


def extra_files_line(dir_: str):
    return dir_, [os.path.join(dir_, i) for i in os.listdir(dir_) if os.path.isfile(os.path.join(dir_, i))]


data_files = [
    extra_files_line('share/applications'),
    extra_files_line('share/eduvpn'),
    extra_files_line('share/eduvpn/builder'),
    extra_files_line('share/eduvpn/images'),
    extra_files_line('share/eduvpn/images/flags'),
    extra_files_line('share/letsconnect/images'),
    extra_files_line('share/icons/hicolor/48x48/apps'),
    extra_files_line('share/icons/hicolor/128x128/apps'),
    extra_files_line('share/icons/hicolor/256x256/apps'),
    extra_files_line('share/icons/hicolor/512x512/apps'),
]
for dir in glob('share/locale/*/LC_MESSAGES'):
    data_files.append([dir, glob(os.path.join(dir,'*.mo'))])

setup(
    name="eduvpn_client",
    version=__version__,
    packages=find_packages(exclude=['tests']),
    data_files=data_files,
    install_requires=install_requires,
    extras_require=extras_require,
    author="Gijs Molenaar",
    author_email="gijs@pythonic.nl",
    description="eduVPN client",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="GPL3",
    setup_requires=['pytest-runner'],
    tests_require=tests_require,
    test_suite="tests",
    keywords="vpn openvpn networking security",
    url="https://github.com/eduvpn/python-eduvpn-client",
    python_requires='>=3.6',
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: System :: Operating System Kernels :: Linux",
        "Topic :: System :: Networking",
        "Environment :: X11 Applications",
    ],
    entry_points={
        'console_scripts': [
            'eduvpn-cli = eduvpn.cli:eduvpn',
            'letsconnect-cli = eduvpn.cli:letsconnect',
        ],
        'gui_scripts': [
            'eduvpn-gui = eduvpn.ui.__main__:eduvpn',
            'letsconnect-gui = eduvpn.ui.__main__:letsconnect',
        ]
    }
)



from typing import Optional, Iterable, Callable, List, Dict
import enum
from functools import lru_cache
from gi.repository import Gtk, GObject, Pango
from eduvpn.server import (
    AnyServer as Server, InstituteAccessServer,
    OrganisationServer, SecureInternetLocation, CustomServer)
from .utils import show_ui_component


class ServerGroup(enum.Enum):
    INSTITUTE_ACCESS = enum.auto()
    SECURE_INTERNET = enum.auto()
    OTHER = enum.auto()


group_tree_component = {
    ServerGroup.INSTITUTE_ACCESS: 'institute_list',
    ServerGroup.SECURE_INTERNET: 'secure_internet_list',
    ServerGroup.OTHER: 'other_server_list',
}

group_header_component = {
    ServerGroup.INSTITUTE_ACCESS: 'institute_list_header',
    ServerGroup.SECURE_INTERNET: 'secure_internet_list_header',
    ServerGroup.OTHER: 'other_server_list_header',
}


# typing aliases
Model = Gtk.ListStore


@lru_cache()
def get_group_model(group: ServerGroup) -> Model:
    # Model: (name: str, server: ServerType)
    return Gtk.ListStore(  # type: ignore
        GObject.TYPE_STRING,
        GObject.TYPE_PYOBJECT)


def server_to_model_data(server: Server) -> list:
    return [str(server), server]


def show_result_components(window, show: bool):
    """
    Set the visibility of essential server list related components.
    """
    window.show_page(window.find_server_page)
    show_ui_component(window.institute_list, show)
    show_ui_component(window.secure_internet_list, show)
    show_ui_component(window.other_server_list, show)


def show_search_components(window, show: bool):
    """
    Set the visibility of essential search related components.
    """
    show_ui_component(window.find_server_search_form, show)
    show_ui_component(window.find_server_image, show)
    show_ui_component(window.find_server_label, show)
    show_ui_component(window.find_server_search_input, show)


def show_search_results(window, show: bool):
    """
    Set the visibility of the tree of the search result component in the UI.
    """
    show_ui_component(window.server_list_container, show)


def group_servers(
        servers: Iterable[Server]) -> Dict[ServerGroup, List[Server]]:
    """
    Separate the servers into three groups.
    """
    groups: Dict[ServerGroup, List[Server]] = {
        ServerGroup.INSTITUTE_ACCESS: [],
        ServerGroup.SECURE_INTERNET: [],
        ServerGroup.OTHER: [],
    }
    for server in servers:
        if isinstance(server, InstituteAccessServer):
            groups[ServerGroup.INSTITUTE_ACCESS].append(server)
        elif isinstance(server, (OrganisationServer, SecureInternetLocation)):
            groups[ServerGroup.SECURE_INTERNET].append(server)
        elif isinstance(server, CustomServer):
            groups[ServerGroup.OTHER].append(server)
        else:
            raise TypeError(server)
    return groups


def show_group_tree(window, group: ServerGroup, show: bool):
    """
    Set the visibility of the tree of result for a server type.
    """
    tree_component_name = group_tree_component[group]
    tree_component = getattr(window, tree_component_name)
    show_ui_component(tree_component, show)
    header_component_name = group_header_component[group]
    header_component = getattr(window, header_component_name)
    show_ui_component(header_component, show)


def init_server_search(window):
    "Initialize the search page components."
    text_cell = Gtk.CellRendererText()
    text_cell.props.ellipsize = Pango.EllipsizeMode.END
    text_cell.set_property("size-points", 14)
    for group in group_tree_component:
        component_name = group_tree_component[group]
        tree_view = getattr(window, component_name)
        if len(tree_view.get_columns()) == 0:
            # Only add this column once.
            column = Gtk.TreeViewColumn(None, text_cell, text=0)
            tree_view.append_column(column)
        model = get_group_model(group)
        tree_view.set_model(model)


def exit_server_search(window):
    "Hide the search page components."
    for group in group_tree_component:
        show_group_tree(window, group, False)
    show_search_results(window, False)


def connect_selection_handlers(window, select_callback: Callable):
    "Connect the selection callback handlers for each server type."
    for group in group_tree_component:
        component_name = group_tree_component[group]
        tree_view = getattr(window, component_name)
        selection = tree_view.get_selection()
        selection.connect("changed", select_callback)


def disconnect_selection_handlers(window, select_callback: Callable):
    "Disconnect the selection callback handlers for each server type."
    for group in group_tree_component:
        component_name = group_tree_component[group]
        tree_view = getattr(window, component_name)
        selection = tree_view.get_selection()
        selection.disconnect_by_func(select_callback)


def update_search_results_for_type(window,
                                   group: ServerGroup,
                                   servers: Iterable[Server]):
    """
    Update the UI with the search results
    for a single type of server.
    """
    model = get_group_model(group)  # type: ignore
    # Remove the old search results.
    model.clear()  # type: ignore
    # Add the new search results.
    for server in servers:
        model_data = server_to_model_data(server)
        model.append(model_data)  # type: ignore
    # Update the UI.
    model_has_results = len(model) > 0  # type: ignore
    show_group_tree(window, group, show=model_has_results)


def update_results(window, servers: Optional[Iterable[Server]]):
    """
    Update the UI with the search results.
    """
    if servers is None:
        show_search_results(window, False)
        return
    server_map = group_servers(servers)
    for group in group_tree_component:
        update_search_results_for_type(
            window,
            group,
            server_map.get(group, []),
        )
    show_search_results(window, True)



import logging
from pathlib import Path
from typing import Optional, TextIO

from ..nm import get_iface, get_ipv4, get_ipv6
from ..utils import cache, get_human_readable_bytes, translated_property

logger = logging.getLogger(__name__)

LINUX_NET_FOLDER = Path("/sys/class/net")


class NetworkStats:

    default_text = translated_property("N/A")

    # These properties define an LRU cache
    # This cache is used so that we do not query these every second
    @property  # type: ignore
    @cache
    def ipv4(self) -> str:
        _ipv4 = get_ipv4()
        if _ipv4 is None:
            _ipv4 = self.default_text
        return _ipv4

    @property  # type: ignore
    @cache
    def ipv6(self) -> str:
        _ipv6 = get_ipv6()
        if _ipv6 is None:
            _ipv6 = self.default_text
        return _ipv6

    @property  # type: ignore
    @cache
    def upload_file(self) -> Optional[TextIO]:
        return self.open_file("tx_bytes")

    @property  # type: ignore
    @cache
    def download_file(self) -> Optional[TextIO]:
        return self.open_file("rx_bytes")

    @property  # type: ignore
    @cache
    def start_bytes_upload(self) -> Optional[int]:
        return self.get_file_bytes(self.upload_file)  # type: ignore

    @property  # type: ignore
    @cache
    def start_bytes_download(self) -> Optional[int]:
        return self.get_file_bytes(self.download_file)  # type: ignore

    @property  # type: ignore
    @cache
    def iface(self) -> Optional[str]:
        return get_iface()

    @property
    def download(self) -> str:
        """
        Get the download as a human readable string
        """
        file_bytes_download = self.get_file_bytes(self.download_file)  # type: ignore
        if file_bytes_download is None:
            return self.default_text
        if file_bytes_download <= self.start_bytes_download:  # type: ignore
            return get_human_readable_bytes(0)
        return get_human_readable_bytes(file_bytes_download - self.start_bytes_download)  # type: ignore

    @property
    def upload(self) -> str:
        """
        Get the upload as a human readable string
        """
        file_bytes_upload = self.get_file_bytes(self.upload_file)  # type: ignore
        if file_bytes_upload is None:
            return self.default_text
        if file_bytes_upload <= self.start_bytes_upload:  # type: ignore
            return get_human_readable_bytes(0)
        return get_human_readable_bytes(file_bytes_upload - self.start_bytes_upload)  # type:ignore

    def open_file(self, filename: str) -> Optional[TextIO]:
        """
        Helper function to open a statistics network file
        """
        if not self.iface:
            logger.warning(f"Network Stats: {filename}, failed to get interface")
            return None
        filepath = LINUX_NET_FOLDER / self.iface / "statistics" / filename  # type: ignore
        if not filepath.is_file():
            logger.warning(f"Network Stats: {filepath} is not a file")
            return None
        return open(filepath, "r")

    def get_file_bytes(self, filehandler: Optional[TextIO]) -> Optional[int]:
        """
        Helper function to get a statistics file to calculate the total data transfer
        """
        # If the interface is not set
        # or the file is not present, we cannot get the stat
        if not self.iface:
            # Warning was already shown
            return None
        if not filehandler:
            # Warning was already shown
            return None

        # Get the statistic from the file
        # and go to the beginning
        try:
            stat = int(filehandler.readline())
        except ValueError:
            stat = 0
        filehandler.seek(0)
        return stat

    def cleanup(self):
        """
        Cleanup the network stats by closing the files
        """
        if self.download_file:
            self.download_file.close()
        if self.upload_file:
            self.upload_file.close()



import logging
from typing import Optional, Callable
from functools import partial
from http.server import BaseHTTPRequestHandler, HTTPServer
from urllib.parse import urlparse, parse_qs
from gettext import gettext as _
import requests
from ..variants import ApplicationVariant
from ..utils import run_in_background_thread


logger = logging.getLogger(__name__)


HTTP_HOST = 'localhost'
CALLBACK_PATH = '/callback'
CANCEL_PATH = '/cancel'


landing_page = """
<!doctype html>
<html lang=en>
<head>
<meta charset=utf-8>
<title>{brand}</title>
<style>
.center {{
    font-family: arial;
    font-size: 30px;
    position: absolute;
    text-align: center;
    width: 800px;
    height: 50px;
    top: 50%;
    left: 50%;
    margin-left: -400px; /* margin is -0.5 * dimension */
    margin-top: -25px;
}}
</style>
</head>
<body>

<div class="center">
<img src="data:image/png;base64,{logo}" width="150">
<p>{message}</p>
</div>
</body>
</html>
"""


def stringify_image(logo: str) -> str:
    import base64
    with open(logo, 'rb') as logo_file:
        return base64.b64encode(logo_file.read()).decode('ascii')


def build_response_page(app_variant) -> bytes:
    logo = stringify_image(app_variant.icon)
    message = _("You can now close this window.")
    return landing_page.format(logo=logo, brand=app_variant.name, message=message).encode('utf-8')


class OAuthWebServer:
    scheme = 'http'

    def __init__(self, app_variant: ApplicationVariant):
        self.app_variant = app_variant
        handler = partial(RequestHandler, oauth_web_server=self)
        self.server = HTTPServer((HTTP_HOST, 0), handler)
        self._completed = False
        self._result = None

    @property
    def address(self):
        return f'{self.server.server_address[0]}:{self.server.server_port}'

    @property
    def success_url(self):
        return f'{self.scheme}://{self.address}{CALLBACK_PATH}'

    @property
    def cancel_url(self):
        return f'{self.scheme}://{self.address}{CANCEL_PATH}'

    def run(self) -> Optional[dict]:
        logger.info(f"listening for a request at {self.address}...")
        self.server.handle_request()
        self.server.server_close()
        if not self._completed:
            logger.error("invalid request received")
        return self._result

    @run_in_background_thread('oauth-http-server-run')
    def run_in_background(self, callback: Callable[[Optional[dict]], None]):
        session = self.run()
        callback(session)

    @run_in_background_thread('oauth-http-server-stop')
    def stop(self):
        requests.get(self.cancel_url)


class RequestHandler(BaseHTTPRequestHandler):
    def __init__(self, *args, oauth_web_server: OAuthWebServer, **kwargs):
        self._oauth_web_server = oauth_web_server
        super().__init__(*args, **kwargs)

    def do_GET(self):
        logger.info(f"received a request {self.path}")
        # notify the auth webserver of the result
        self._oauth_web_server._completed = True
        if self.path == CANCEL_PATH:
            self._oauth_web_server._result = None
        else:
            url = urlparse(self.path)
            query = parse_qs(url.query)
            self._oauth_web_server._result = query
        # send back a response
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        page = build_response_page(self._oauth_web_server.app_variant)
        self.wfile.write(page)



import logging
import os
import sys

import gi
gi.require_version('Gtk', '3.0')  # noqa: E402
from gi.repository import Gtk

from ..variants import EDUVPN, LETS_CONNECT
from ..ui.app import EduVpnGtkApplication


logger = logging.getLogger(__name__)


def main_loop(args=None, app_variant=EDUVPN):
    if os.geteuid() == 0:
        print("Running client as root is not supported")
        sys.exit(1)

    if args is None:
        args = sys.argv

    try:
        app = EduVpnGtkApplication(app_variant=app_variant)
        app.run(args)
    except Exception as e:
        message = "A fatal error occurred"
        logger.exception(message)
        dialog = Gtk.MessageDialog(
            text=f"{message}: {e!r}",
            message_type=Gtk.MessageType.ERROR,
            modal=True,
            buttons=Gtk.ButtonsType.OK,
        )
        dialog.connect('response', Gtk.main_quit)
        dialog.show()
        Gtk.main()
    except KeyboardInterrupt:
        pass


def eduvpn(args=None):
    main_loop(args, app_variant=EDUVPN)


def letsconnect(args=None):
    main_loop(args, app_variant=LETS_CONNECT)


if __name__ == '__main__':
    eduvpn()



from unittest import TestCase
from eduvpn.i18n import extract_translation


class TestI18n(TestCase):
    def test_extract_translation(self):
        self.assertEqual("bla bla", extract_translation("bla bla"))

        self.assertEqual('bleu bleu', extract_translation({'de': 'bleu bleu'}))



from typing import Optional
from . import settings


class ApplicationVariant:
    def __init__(
        self,
        app_id: str,
        name: str,
        icon: str,
        translation_domain: str,
        logo: Optional[str] = None,
        server_image: Optional[str] = None,
        use_predefined_servers: bool = True,
        use_configured_servers: bool = True,
    ):
        self.app_id = app_id
        self.name = name
        self.icon = icon
        self.logo = logo
        self.server_image = server_image
        self.translation_domain = translation_domain
        self.use_predefined_servers = use_predefined_servers
        self.use_configured_servers = use_configured_servers


EDUVPN = ApplicationVariant(
    app_id='org.eduvpn.client',
    name=settings.EDUVPN_NAME,
    icon=settings.EDUVPN_ICON,
    translation_domain='eduVPN',
)

LETS_CONNECT = ApplicationVariant(
    app_id='org.letsconnect-vpn.client',
    name=settings.LETS_CONNECT_NAME,
    icon=settings.LETS_CONNECT_ICON,
    logo=settings.LETS_CONNECT_LOGO,
    server_image=settings.SERVER_ILLUSTRATION,
    translation_domain='LetConnect',
    use_predefined_servers=False,
    use_configured_servers=False,
)



from unittest import TestCase
from eduvpn.ovpn import InvalidOVPN, Ovpn, Field, Section, Comment, Empty


class OVPNTests(TestCase):
    def test_parse_and_write(self):
        ovpn_file = (
            '# hello world\n'
            'set a 1\n'
            '\n'
            '<ca>\n'
            'the\n'
            'certificate\n'
            'content\n'
            '</ca>\n'
        )
        ovpn = Ovpn.parse(ovpn_file)
        self.assertEqual(ovpn.content, [
            Comment(' hello world'),
            Field('set', ['a', '1']),
            Empty(),
            Section('ca', ['the', 'certificate', 'content']),
        ])
        self.assertEqual(ovpn.to_string(), ovpn_file)

    def test_force_tcp(self):
        ovpn = Ovpn.parse(
            'a\n'
            'remote vpn.example.com 1234 udp\n'
            'remote vpn.example.com 5678 udp\r\n'
            'remote vpn.example.com 1234 tcp\r'
            'remote vpn.example.com 5678 tcp\r\n'
            'b\r\n'
            'c\n'
        )
        expected_content = (
            'a \n'
            '# omitted to force tcp: remote vpn.example.com 1234 udp\n'
            '# omitted to force tcp: remote vpn.example.com 5678 udp\n'
            'remote vpn.example.com 1234 tcp\n'
            'remote vpn.example.com 5678 tcp\n'
            'b \n'
            'c \n'
        )
        ovpn.force_tcp()
        self.assertEqual(ovpn.to_string(), expected_content)

    def test_force_tcp_fail(self):
        ovpn = Ovpn.parse(
            'remote vpn.example.com 1234 udp\n'
        )
        with self.assertRaises(InvalidOVPN):
            ovpn.force_tcp()



from typing import TypeVar, Generic, Any, Dict
import logging
import json
from .settings import CONFIG_PREFIX


T = TypeVar('T')


CONFIG_FILE_NAME = 'config.json'

CONFIG_PATH = CONFIG_PREFIX / CONFIG_FILE_NAME

DEFAULT_SETTINGS = dict(
    force_tcp=False,
)


logger = logging.getLogger(__name__)


class SettingDescriptor(Generic[T]):
    def __set_name__(self, owner, name):
        self.name = name

    def __get__(self, instance, owner=None) -> T:
        return instance.get_setting(self.name)

    def __set__(self, instance, value: T):
        instance.set_setting(self.name, value)


class Configuration:
    def __init__(self, settings: Dict[str, Any]):
        self.settings = settings

    @classmethod
    def load(cls):
        if not CONFIG_PATH.exists():
            return cls(dict(DEFAULT_SETTINGS))
        with open(CONFIG_PATH, 'r') as f:
            try:
                settings = json.load(f)
            except Exception:
                logger.exception("error loading settings")
                settings = {}
            else:
                logger.debug(f"loaded settings: {settings}")
        settings = {**DEFAULT_SETTINGS, **settings}
        return cls(settings)

    def save(self):
        logger.debug(f"saving settings: {self.settings}")
        with open(CONFIG_PATH, 'w') as f:
            json.dump(self.settings, f)

    def get_setting(self, name: str):
        return self.settings[name]

    def set_setting(self, name: str, value: Any):
        if value != self.settings[name]:
            self.settings[name] = value
            self.save()

    force_tcp = SettingDescriptor[bool]()
